@@ -1,6 +1,6 @@
 // See the file "COPYING" in the main distribution directory for copyright.
 
-#include "config.h"
+#include "bro-config.h"
 
 #include <algorithm>
 #include <ctype.h>
@@ -166,17 +166,19 @@ void BroString::Set(const BroString& str)
 
 const char* BroString::CheckString() const
 	{
+	void *nulTerm;
 	if ( n == 0 )
 		return "";
 
-	if ( memchr(b, '\0', n + final_NUL) != &b[n] )
+	nulTerm = memchr(b, '\0', n + final_NUL);
+	if ( nulTerm != &b[n] )
 		{
 		// Either an embedded NUL, or no final NUL.
 		char* exp_s = Render();
-		if ( b[n-1] != '\0' )
-			reporter->Error("string without NUL terminator: \"%s\"", exp_s);
-		else
+		if ( nulTerm )
 			reporter->Error("string with embedded NUL: \"%s\"", exp_s);
+		else
+			reporter->Error("string without NUL terminator: \"%s\"", exp_s);
 
 		delete [] exp_s;
 		return "<string-with-NUL>";
@@ -194,22 +196,7 @@ char* BroString::Render(int format, int* len) const
 
 	for ( int i = 0; i < n; ++i )
 		{
-		if ( b[i] == '\0' && (format & ESC_NULL) )
-			{
-			*sp++ = '\\'; *sp++ = '0';
-			}
-
-		else if ( b[i] == '\x7f' && (format & ESC_DEL) )
-			{
-			*sp++ = '^'; *sp++ = '?';
-			}
-
-		else if ( b[i] <= 26 && (format & ESC_LOW) )
-			{
-			*sp++ = '^'; *sp++ = b[i] + 'A' - 1;
-			}
-
-		else if ( b[i] == '\\' && (format & ESC_ESC) )
+		if ( b[i] == '\\' && (format & ESC_ESC) )
 			{
 			*sp++ = '\\'; *sp++ = '\\';
 			}
