@@ -19,6 +19,7 @@
 #include "Conn.h"
 #include "Timer.h"
 #include "RemoteSerializer.h"
+#include "iosource/Manager.h"
 
 Serializer::Serializer(SerializationFormat* arg_format)
 	{
@@ -81,6 +82,7 @@ bool Serializer::EndSerialization(SerialInfo* info)
 
 	ChunkedIO::Chunk* chunk = new ChunkedIO::Chunk;
 	chunk->len = format->EndWrite(&chunk->data);
+	chunk->free_func = ChunkedIO::Chunk::free_func_free;
 
 	if ( ! io->Write(chunk) )
 		{
@@ -282,7 +284,6 @@ int Serializer::Unserialize(UnserialInfo* info, bool block)
 
 	if ( ! info->chunk )
 		{ // only delete if we allocated it ourselves
-		delete [] chunk->data;
 		delete chunk;
 		}
 
@@ -436,7 +437,7 @@ bool Serializer::UnserializeCall(UnserialInfo* info)
 
 bool Serializer::UnserializeStateAccess(UnserialInfo* info)
 	{
-	SetErrorDescr("unserializing state acess");
+	SetErrorDescr("unserializing state access");
 
 	StateAccess* s = StateAccess::Unserialize(info);
 
@@ -975,69 +976,6 @@ void FileSerializer::GotPacket(Packet* p)
 	delete p;
 	}
 
-ConversionSerializer::ConversionSerializer(SerializationFormat* in,
-						SerializationFormat* out)
-: FileSerializer(in)
-	{
-	serout = new FileSerializer(out);
-	}
-
-ConversionSerializer::~ConversionSerializer()
-	{
-	delete serout;
-	}
-
-bool ConversionSerializer::Convert(const char* file_in, const char* file_out)
-	{
-	reporter->InternalError("Error: Printing as XML is broken.");
-
-	if ( ! serout->Open(file_out, true) )
-		return false;
-
-	UnserialInfo info_in(this);
-	if ( ! Read(&info_in, file_in) )
-		return false;
-
-	if ( ! serout->Close() )
-		return false;
-
-	return true;
-	}
-
-void ConversionSerializer::GotEvent(const char* name, double time,
-					EventHandlerPtr event, val_list* args)
-	{
-	SerialInfo info(serout);
-	serout->Serialize(&info, name, args);
-	delete_vals(args);
-	}
-
-void ConversionSerializer::GotFunctionCall(const char* name, double time,
-					Func* func, val_list* args)
-	{
-	SerialInfo info(serout);
-	serout->Serialize(&info, name, args);
-	delete_vals(args);
-	}
-
-void ConversionSerializer::GotID(ID* id, Val* val)
-	{
-	reporter->Warning("ConversionSerializer::GotID not implemented");
-	Unref(id);
-	}
-
-void ConversionSerializer::GotStateAccess(StateAccess* s)
-	{
-	reporter->Warning("ConversionSerializer::GotID not implemented");
-	delete s;
-	}
-
-void ConversionSerializer::GotPacket(Packet* p)
-	{
-	reporter->Warning("ConversionSerializer::GotPacket not implemented");
-	delete p;
-	}
-
 EventPlayer::EventPlayer(const char* file)
     : stream_time(), replay_time(), ne_time(), ne_handler(), ne_args()
 	{
@@ -1045,7 +983,7 @@ EventPlayer::EventPlayer(const char* file)
 		Error(fmt("event replayer: cannot open %s", file));
 
 	if ( ReadHeader() )
-		io_sources.Register(this);
+		iosource_mgr->Register(this);
 	}
 
 EventPlayer::~EventPlayer()
@@ -1067,9 +1005,10 @@ void EventPlayer::GotFunctionCall(const char* name, double time,
 	// We don't replay function calls.
 	}
 
-void EventPlayer::GetFds(int* read, int* write, int* except)
+void EventPlayer::GetFds(iosource::FD_Set* read, iosource::FD_Set* write,
+                         iosource::FD_Set* except)
 	{
-	*read = fd;
+	read->Insert(fd);
 	}
 
 double EventPlayer::NextTimestamp(double* local_network_time)
@@ -1078,18 +1017,18 @@ double EventPlayer::NextTimestamp(double* local_network_time)
 		return ne_time;
 
 	if ( ! io )
-		return 0;
+		return -1;
 
 	// Read next event if we don't have one waiting.
 	if ( ! ne_time )
 		{
 		UnserialInfo info(this);
 		Unserialize(&info);
-		closed = io->Eof();
+		SetClosed(io->Eof());
 		}
 
 	if ( ! ne_time )
-		return 0;
+		return -1;
 
 	if ( ! network_time )
 		{
@@ -1120,104 +1059,3 @@ void EventPlayer::Process()
 
 	ne_time = 0;
 	}
-
-void Packet::Describe(ODesc* d) const
-	{
-	const IP_Hdr ip = IP();
-	d->Add(ip.SrcAddr());
-	d->Add("->");
-	d->Add(ip.DstAddr());
-	}
-
-bool Packet::Serialize(SerialInfo* info) const
-	{
-	return SERIALIZE(uint32(hdr->ts.tv_sec)) &&
-		SERIALIZE(uint32(hdr->ts.tv_usec)) &&
-		SERIALIZE(uint32(hdr->len)) &&
-		SERIALIZE(link_type) &&
-		info->s->Write(tag.c_str(), 0, "tag") &&
-		info->s->Write((const char*) pkt, hdr->caplen, "data");
-	}
-
-static BroFile* profiling_output = 0;
-
-#ifdef DEBUG
-static PktDumper* dump = 0;
-#endif
-
-Packet* Packet::Unserialize(UnserialInfo* info)
-	{
-	Packet* p = new Packet("", true);
-	pcap_pkthdr* hdr = new pcap_pkthdr;
-
-	uint32 tv_sec, tv_usec, len;
-
-	if ( ! (UNSERIALIZE(&tv_sec) &&
-		UNSERIALIZE(&tv_usec) &&
-		UNSERIALIZE(&len) &&
-		UNSERIALIZE(&p->link_type)) )
-		{
-		delete p;
-		delete hdr;
-		return 0;
-		}
-
-	hdr->ts.tv_sec = tv_sec;
-	hdr->ts.tv_usec = tv_usec;
-	hdr->len = len;
-
-	char* tag;
-	if ( ! info->s->Read((char**) &tag, 0, "tag") )
-		{
-		delete p;
-		delete hdr;
-		return 0;
-		}
-
-	char* pkt;
-	int caplen;
-	if ( ! info->s->Read((char**) &pkt, &caplen, "data") )
-		{
-		delete p;
-		delete hdr;
-		delete [] tag;
-		return 0;
-		}
-
-	hdr->caplen = uint32(caplen);
-	p->hdr = hdr;
-	p->pkt = (u_char*) pkt;
-	p->tag = tag;
-	p->hdr_size = get_link_header_size(p->link_type);
-
-	delete [] tag;
-
-	// For the global timer manager, we take the global network_time as the
-	// packet's timestamp for feeding it into our packet loop.
-	if ( p->tag == "" )
-		p->time = timer_mgr->Time();
-	else
-		p->time = p->hdr->ts.tv_sec + double(p->hdr->ts.tv_usec) / 1e6;
-
-	if ( time_machine_profiling )
-		{
-		if ( ! profiling_output )
-			profiling_output =
-				new BroFile("tm-prof.packets.log", "w");
-
-		profiling_output->Write(fmt("%.6f %s %d\n", current_time(),
-			(p->tag != "" ? p->tag.c_str() : "-"), hdr->len));
-		}
-
-#ifdef DEBUG
-	if ( debug_logger.IsEnabled(DBG_TM) )
-		{
-		if ( ! dump )
-			dump = new PktDumper("tm.pcap");
-
-		dump->Dump(p->hdr, p->pkt);
-		}
-#endif
-
-	return p;
-	}
