@@ -16,17 +16,17 @@ namespace analyzer { namespace socks {
 
 class SOCKS_Analyzer : public tcp::TCP_ApplicationAnalyzer {
 public:
-	SOCKS_Analyzer(Connection* conn);
-	~SOCKS_Analyzer();
+	explicit SOCKS_Analyzer(Connection* conn);
+	~SOCKS_Analyzer() override;
 
 	void EndpointDone(bool orig);
 
-	virtual void Done();
-	virtual void DeliverStream(int len, const u_char* data, bool orig);
-	virtual void Undelivered(uint64 seq, int len, bool orig);
-	virtual void EndpointEOF(bool is_orig);
+	void Done() override;
+	void DeliverStream(int len, const u_char* data, bool orig) override;
+	void Undelivered(uint64 seq, int len, bool orig) override;
+	void EndpointEOF(bool is_orig) override;
 
-	static analyzer::Analyzer* InstantiateAnalyzer(Connection* conn)
+	static analyzer::Analyzer* Instantiate(Connection* conn)
 		{ return new SOCKS_Analyzer(conn); }
 
 protected:
